=== modified file 'NEWS'
--- NEWS	2008-11-25 20:35:28 +0000
+++ NEWS	2008-11-26 09:05:22 +0000
@@ -47,6 +47,10 @@
       "PermissionDenied: "None"" on the client.
       (Andrew Bennetts, #299254)
       
+    * Don't crash when requesting log --forward <file> for a revision range
+      starting with a dotted revno.
+      (Vincent Ladeuil, #300055)
+
     * TooManyConcurrentRequests no longer occur when a fetch fails and
       tries to abort a write group.  This allows the root cause (e.g. a
       network interruption) to be reported.  (Andrew Bennetts, #297014)

=== modified file 'bzrlib/log.py'
--- bzrlib/log.py	2008-10-01 05:40:45 +0000
+++ bzrlib/log.py	2008-11-25 08:25:20 +0000
@@ -236,18 +236,17 @@
 def calculate_view_revisions(branch, start_revision, end_revision, direction,
                              specific_fileid, generate_merge_revisions,
                              allow_single_merge_revision):
-    if (not generate_merge_revisions and start_revision is end_revision is
-        None and direction == 'reverse' and specific_fileid is None):
+    if (    not generate_merge_revisions
+        and start_revision is end_revision is None
+        and direction == 'reverse'
+        and specific_fileid is None):
         return _linear_view_revisions(branch)
 
-    mainline_revs, rev_nos, start_rev_id, end_rev_id = \
-        _get_mainline_revs(branch, start_revision, end_revision)
+    mainline_revs, rev_nos, start_rev_id, end_rev_id = _get_mainline_revs(
+        branch, start_revision, end_revision)
     if not mainline_revs:
         return []
 
-    if direction == 'reverse':
-        start_rev_id, end_rev_id = end_rev_id, start_rev_id
-
     generate_single_revision = False
     if ((not generate_merge_revisions)
         and ((start_rev_id and (start_rev_id not in rev_nos))
@@ -260,6 +259,9 @@
         generate_merge_revisions = generate_single_revision
     view_revs_iter = get_view_revisions(mainline_revs, rev_nos, branch,
                           direction, include_merges=generate_merge_revisions)
+
+    if direction == 'reverse':
+        start_rev_id, end_rev_id = end_rev_id, start_rev_id
     view_revisions = _filter_revision_range(list(view_revs_iter),
                                             start_rev_id,
                                             end_rev_id)
@@ -267,9 +269,8 @@
         view_revisions = view_revisions[0:1]
     if specific_fileid:
         view_revisions = _filter_revisions_touching_file_id(branch,
-                                                         specific_fileid,
-                                                         view_revisions,
-                                                         direction)
+                                                            specific_fileid,
+                                                            view_revisions)
 
     # rebase merge_depth - unless there are no revisions or 
     # either the first or last revision have merge_depth = 0.
@@ -438,7 +439,7 @@
     # filtered later.
     # Also map the revisions to rev_ids, to be used in the later filtering
     # stage.
-    start_rev_id = None 
+    start_rev_id = None
     if start_revision is None:
         start_revno = 1
     else:
@@ -448,7 +449,7 @@
         else:
             branch.check_real_revno(start_revision)
             start_revno = start_revision
-    
+
     end_rev_id = None
     if end_revision is None:
         end_revno = branch_revno
@@ -507,7 +508,7 @@
 
     :return: The filtered view_revisions.
     """
-    if start_rev_id or end_rev_id: 
+    if start_rev_id or end_rev_id:
         revision_ids = [r for r, n, d in view_revisions]
         if start_rev_id:
             start_index = revision_ids.index(start_rev_id)
@@ -536,8 +537,7 @@
     return view_revisions
 
 
-def _filter_revisions_touching_file_id(branch, file_id, view_revisions,
-                                       direction):
+def _filter_revisions_touching_file_id(branch, file_id, view_revisions):
     r"""Return the list of revision ids which touch a given file id.
 
     The function filters view_revisions and returns a subset.
@@ -562,13 +562,14 @@
     This will also be restricted based on a subset of the mainline.
 
     :param branch: The branch where we can get text revision information.
+
     :param file_id: Filter out revisions that do not touch file_id.
+
     :param view_revisions: A list of (revision_id, dotted_revno, merge_depth)
         tuples. This is the list of revisions which will be filtered. It is
-        assumed that view_revisions is in merge_sort order (either forward or
-        reverse).
-    :param direction: The direction of view_revisions.  See also
-        reverse_by_depth, and get_view_revisions
+        assumed that view_revisions is in merge_sort order (i.e. newest
+        revision first ).
+
     :return: A list of (revision_id, dotted_revno, merge_depth) tuples.
     """
     # Lookup all possible text keys to determine which ones actually modified
@@ -592,11 +593,6 @@
     del text_keys, next_keys
 
     result = []
-    if direction == 'forward':
-        # TODO: The algorithm for finding 'merges' of file changes expects
-        #       'reverse' order (the default from 'merge_sort()'). Instead of
-        #       forcing this, we could just use the reverse_by_depth order.
-        view_revisions = reverse_by_depth(view_revisions)
     # Track what revisions will merge the current revision, replace entries
     # with 'None' when they have been added to result
     current_merge_stack = [None]
@@ -615,8 +611,6 @@
                 if node is not None:
                     result.append(node)
                     current_merge_stack[idx] = None
-    if direction == 'forward':
-        result = reverse_by_depth(result)
     return result
 
 
@@ -666,19 +660,27 @@
     revision of that depth.  There may be no topological justification for this,
     but it looks much nicer.
     """
+    # Add a fake revision at start so that we can always attach sub revisions
+    merge_sorted_revisions = [(None, None, _depth)] + merge_sorted_revisions
     zd_revisions = []
     for val in merge_sorted_revisions:
         if val[2] == _depth:
+            # Each revision at the current depth becomes a chunk grouping all
+            # higher depth revisions.
             zd_revisions.append([val])
         else:
             zd_revisions[-1].append(val)
     for revisions in zd_revisions:
         if len(revisions) > 1:
+            # We have higher depth revisions, let reverse them locally
             revisions[1:] = reverse_by_depth(revisions[1:], _depth + 1)
     zd_revisions.reverse()
     result = []
     for chunk in zd_revisions:
         result.extend(chunk)
+    if _depth == 0:
+        # Top level call, get rid of the fake revisions that have been added
+        result = [r for r in result if r[0] is not None and r[1] is not None]
     return result
 
 
@@ -861,7 +863,7 @@
         return str[:max_len-3]+'...'
 
     def date_string(self, rev):
-        return format_date(rev.timestamp, rev.timezone or 0, 
+        return format_date(rev.timestamp, rev.timezone or 0,
                            self.show_timezone, date_fmt="%Y-%m-%d",
                            show_offset=False)
 

=== modified file 'bzrlib/tests/blackbox/test_log.py'
--- bzrlib/tests/blackbox/test_log.py	2008-10-01 05:40:45 +0000
+++ bzrlib/tests/blackbox/test_log.py	2008-11-21 16:43:53 +0000
@@ -29,7 +29,8 @@
 from bzrlib.tests import test_log
 
 
-class TestCaseWithoutPropsHandler(ExternalBase, test_log.TestCaseWithoutPropsHandler):
+class TestCaseWithoutPropsHandler(ExternalBase,
+                                  test_log.TestCaseWithoutPropsHandler):
     pass
 
 
@@ -170,8 +171,8 @@
         branch = tree.branch
         branch.tags.set_tag('tag1', branch.get_rev_id(1))
         branch.tags.set_tag('tag1.1', branch.get_rev_id(1))
-        branch.tags.set_tag('tag3', branch.last_revision()) 
-        
+        branch.tags.set_tag('tag3', branch.last_revision())
+
         log = self.run_bzr("log -r-1")[0]
         self.assertTrue('tags: tag3' in log)
 
@@ -302,7 +303,7 @@
 
     def test_merges_partial_range(self):
         self._prepare()
-        out,err = self.run_bzr('log -r1.1.1..1.1.2')
+        out, err = self.run_bzr('log -r1.1.1..1.1.2')
         self.assertEqual('', err)
         log = normalize_log(out)
         self.assertEqualDiff(log, """\
@@ -339,7 +340,7 @@
         out,err = self.run_bzr('log --short -r1.1.1..1.1.2', retcode=3)
         self.assertContainsRe(err, err_msg)
 
- 
+
 class TestLogEncodings(TestCaseInTempDir):
 
     _mu = u'\xb5'

=== modified file 'bzrlib/tests/test_log.py'
--- bzrlib/tests/test_log.py	2008-11-25 19:56:00 +0000
+++ bzrlib/tests/test_log.py	2008-11-26 09:05:22 +0000
@@ -17,45 +17,32 @@
 import os
 from cStringIO import StringIO
 
-from bzrlib import log, registry
-from bzrlib.tests import TestCase, TestCaseWithTransport
-from bzrlib.log import (show_log,
-                        get_view_revisions,
-                        LogRevision,
-                        LogFormatter,
-                        LongLogFormatter,
-                        ShortLogFormatter,
-                        LineLogFormatter)
-from bzrlib.branch import Branch
-from bzrlib.errors import (
-    BzrCommandError,
-    InvalidRevisionNumber,
-    )
-from bzrlib.revision import Revision
-from bzrlib.revisionspec import (
-    RevisionInfo,
-    RevisionSpec,
-    )
-
-
-class TestCaseWithoutPropsHandler(TestCaseWithTransport):
+from bzrlib import (
+    errors,
+    log,
+    registry,
+    revision,
+    revisionspec,
+    tests,
+    )
+
+
+class TestCaseWithoutPropsHandler(tests.TestCaseWithTransport):
 
     def setUp(self):
         super(TestCaseWithoutPropsHandler, self).setUp()
         # keep a reference to the "current" custom prop. handler registry
-        self.properties_handler_registry = \
-            log.properties_handler_registry
+        self.properties_handler_registry = log.properties_handler_registry
         # clean up the registry in log
         log.properties_handler_registry = registry.Registry()
-        
+
     def _cleanup(self):
         super(TestCaseWithoutPropsHandler, self)._cleanup()
         # restore the custom properties handler registry
-        log.properties_handler_registry = \
-            self.properties_handler_registry
-
-
-class LogCatcher(LogFormatter):
+        log.properties_handler_registry = self.properties_handler_registry
+
+
+class LogCatcher(log.LogFormatter):
     """Pull log messages into list rather than displaying them.
 
     For ease of testing we save log messages here rather than actually
@@ -75,15 +62,26 @@
         self.logs.append(revision)
 
 
-class TestShowLog(TestCaseWithTransport):
+class TestShowLog(tests.TestCaseWithTransport):
 
     def checkDelta(self, delta, **kw):
-        """Check the filenames touched by a delta are as expected."""
+        """Check the filenames touched by a delta are as expected.
+
+        Caller only have to pass in the list of files for each part, all
+        unspecified parts are considered empty (and checked as such).
+        """
         for n in 'added', 'removed', 'renamed', 'modified', 'unchanged':
+            # By default we expect an empty list
             expected = kw.get(n, [])
             # strip out only the path components
             got = [x[0] for x in getattr(delta, n)]
-            self.assertEquals(expected, got)
+            self.assertEqual(expected, got)
+
+    def assertInvalidRevisonNumber(self, br, start, end):
+        lf = LogCatcher()
+        self.assertRaises(errors.InvalidRevisionNumber,
+                          log.show_log, br, lf,
+                          start_revision=start, end_revision=end)
 
     def test_cur_revno(self):
         wt = self.make_branch_and_tree('.')
@@ -91,99 +89,79 @@
 
         lf = LogCatcher()
         wt.commit('empty commit')
-        show_log(b, lf, verbose=True, start_revision=1, end_revision=1)
-        self.assertRaises(InvalidRevisionNumber, show_log, b, lf,
-                          start_revision=2, end_revision=1) 
-        self.assertRaises(InvalidRevisionNumber, show_log, b, lf,
-                          start_revision=1, end_revision=2) 
-        self.assertRaises(InvalidRevisionNumber, show_log, b, lf,
-                          start_revision=0, end_revision=2) 
-        self.assertRaises(InvalidRevisionNumber, show_log, b, lf,
-                          start_revision=1, end_revision=0) 
-        self.assertRaises(InvalidRevisionNumber, show_log, b, lf,
-                          start_revision=-1, end_revision=1) 
-        self.assertRaises(InvalidRevisionNumber, show_log, b, lf,
-                          start_revision=1, end_revision=-1) 
-
-    def test_simple_log(self):
-        eq = self.assertEquals
-        
+        log.show_log(b, lf, verbose=True, start_revision=1, end_revision=1)
+
+        # Since there is a single revision in the branch all the combinations
+        # below should fail.
+        self.assertInvalidRevisonNumber(b, 2, 1)
+        self.assertInvalidRevisonNumber(b, 1, 2)
+        self.assertInvalidRevisonNumber(b, 0, 2)
+        self.assertInvalidRevisonNumber(b, 1, 0)
+        self.assertInvalidRevisonNumber(b, -1, 1)
+        self.assertInvalidRevisonNumber(b, 1, -1)
+
+    def test_empty_branch(self):
         wt = self.make_branch_and_tree('.')
-        b = wt.branch
 
         lf = LogCatcher()
-        show_log(b, lf)
+        log.show_log(wt.branch, lf)
         # no entries yet
-        eq(lf.logs, [])
+        self.assertEqual([], lf.logs)
+
+    def test_empty_commit(self):
+        wt = self.make_branch_and_tree('.')
 
         wt.commit('empty commit')
         lf = LogCatcher()
-        show_log(b, lf, verbose=True)
-        eq(len(lf.logs), 1)
-        eq(lf.logs[0].revno, '1')
-        eq(lf.logs[0].rev.message, 'empty commit')
-        d = lf.logs[0].delta
-        self.log('log delta: %r' % d)
-        self.checkDelta(d)
+        log.show_log(wt.branch, lf, verbose=True)
+        self.assertEqual(1, len(lf.logs))
+        self.assertEqual('1', lf.logs[0].revno)
+        self.assertEqual('empty commit', lf.logs[0].rev.message)
+        self.checkDelta(lf.logs[0].delta)
 
+    def test_simple_commit(self):
+        wt = self.make_branch_and_tree('.')
+        wt.commit('empty commit')
         self.build_tree(['hello'])
         wt.add('hello')
         wt.commit('add one file',
                   committer=u'\u013d\xf3r\xe9m \xcdp\u0161\xfam '
                             u'<test@example.com>')
-
-        lf = self.make_utf8_encoded_stringio()
-        # log using regular thing
-        show_log(b, LongLogFormatter(lf))
-        lf.seek(0)
-        for l in lf.readlines():
-            self.log(l)
-
-        # get log as data structure
         lf = LogCatcher()
-        show_log(b, lf, verbose=True)
-        eq(len(lf.logs), 2)
-        self.log('log entries:')
-        for logentry in lf.logs:
-            self.log('%4s %s' % (logentry.revno, logentry.rev.message))
-
+        log.show_log(wt.branch, lf, verbose=True)
+        self.assertEqual(2, len(lf.logs))
         # first one is most recent
-        logentry = lf.logs[0]
-        eq(logentry.revno, '2')
-        eq(logentry.rev.message, 'add one file')
-        d = logentry.delta
-        self.log('log 2 delta: %r' % d)
-        self.checkDelta(d, added=['hello'])
+        log_entry = lf.logs[0]
+        self.assertEqual('2', log_entry.revno)
+        self.assertEqual('add one file', log_entry.rev.message)
+        self.checkDelta(log_entry.delta, added=['hello'])
 
-        # commit a log message with control characters
+    def test_commit_message_with_control_chars(self):
+        wt = self.make_branch_and_tree('.')
         msg = u"All 8-bit chars: " +  ''.join([unichr(x) for x in range(256)])
         msg = msg.replace(u'\r', u'\n')
-        self.log("original commit message: %r", msg)
         wt.commit(msg)
         lf = LogCatcher()
-        show_log(b, lf, verbose=True)
+        log.show_log(wt.branch, lf, verbose=True)
         committed_msg = lf.logs[0].rev.message
-        self.log("escaped commit message: %r", committed_msg)
-        self.assert_(msg != committed_msg)
-        self.assert_(len(committed_msg) > len(msg))
+        self.assertNotEqual(msg, committed_msg)
+        self.assertTrue(len(committed_msg) > len(msg))
 
-        # Check that log message with only XML-valid characters isn't
+    def test_commit_message_without_control_chars(self):
+        wt = self.make_branch_and_tree('.')
         # escaped.  As ElementTree apparently does some kind of
         # newline conversion, neither LF (\x0A) nor CR (\x0D) are
         # included in the test commit message, even though they are
         # valid XML 1.0 characters.
         msg = "\x09" + ''.join([unichr(x) for x in range(0x20, 256)])
-        self.log("original commit message: %r", msg)
         wt.commit(msg)
         lf = LogCatcher()
-        show_log(b, lf, verbose=True)
+        log.show_log(wt.branch, lf, verbose=True)
         committed_msg = lf.logs[0].rev.message
-        self.log("escaped commit message: %r", committed_msg)
-        self.assert_(msg == committed_msg)
+        self.assertEqual(msg, committed_msg)
 
     def test_deltas_in_merge_revisions(self):
         """Check deltas created for both mainline and merge revisions"""
-        eq = self.assertEquals
         wt = self.make_branch_and_tree('parent')
         self.build_tree(['parent/file1', 'parent/file2', 'parent/file3'])
         wt.add('file1')
@@ -201,56 +179,55 @@
         b = wt.branch
         lf = LogCatcher()
         lf.supports_merge_revisions = True
-        show_log(b, lf, verbose=True)
-        eq(len(lf.logs),3)
+        log.show_log(b, lf, verbose=True)
+
+        self.assertEqual(3, len(lf.logs))
+
         logentry = lf.logs[0]
-        eq(logentry.revno, '2')
-        eq(logentry.rev.message, 'merge child branch')
-        d = logentry.delta
-        self.checkDelta(d, removed=['file1'], modified=['file2'])
+        self.assertEqual('2', logentry.revno)
+        self.assertEqual('merge child branch', logentry.rev.message)
+        self.checkDelta(logentry.delta, removed=['file1'], modified=['file2'])
+
         logentry = lf.logs[1]
-        eq(logentry.revno, '1.1.1')
-        eq(logentry.rev.message, 'remove file1 and modify file2')
-        d = logentry.delta
-        self.checkDelta(d, removed=['file1'], modified=['file2'])
+        self.assertEqual('1.1.1', logentry.revno)
+        self.assertEqual('remove file1 and modify file2', logentry.rev.message)
+        self.checkDelta(logentry.delta, removed=['file1'], modified=['file2'])
+
         logentry = lf.logs[2]
-        eq(logentry.revno, '1')
-        eq(logentry.rev.message, 'add file1 and file2')
-        d = logentry.delta
-        self.checkDelta(d, added=['file1', 'file2'])
+        self.assertEqual('1', logentry.revno)
+        self.assertEqual('add file1 and file2', logentry.rev.message)
+        self.checkDelta(logentry.delta, added=['file1', 'file2'])
 
     def test_merges_nonsupporting_formatter(self):
         """Tests that show_log will raise if the formatter doesn't
         support merge revisions."""
         wt = self.make_branch_and_memory_tree('.')
         wt.lock_write()
-        try:
-            wt.add('')
-            wt.commit('rev-1', rev_id='rev-1',
-                      timestamp=1132586655, timezone=36000,
-                      committer='Joe Foo <joe@foo.com>')
-            wt.commit('rev-merged', rev_id='rev-2a',
-                      timestamp=1132586700, timezone=36000,
-                      committer='Joe Foo <joe@foo.com>')
-            wt.set_parent_ids(['rev-1', 'rev-2a'])
-            wt.branch.set_last_revision_info(1, 'rev-1')
-            wt.commit('rev-2', rev_id='rev-2b',
-                      timestamp=1132586800, timezone=36000,
-                      committer='Joe Foo <joe@foo.com>')
-            logfile = self.make_utf8_encoded_stringio()
-            formatter = ShortLogFormatter(to_file=logfile)
-            wtb = wt.branch
-            lf = LogCatcher()
-            revspec = RevisionSpec.from_string('1.1.1')
-            rev = revspec.in_history(wtb)
-            self.assertRaises(BzrCommandError, show_log, wtb, lf,
-                              start_revision=rev, end_revision=rev)
-        finally:
-            wt.unlock()
+        self.addCleanup(wt.unlock)
+        wt.add('')
+        wt.commit('rev-1', rev_id='rev-1',
+                  timestamp=1132586655, timezone=36000,
+                  committer='Joe Foo <joe@foo.com>')
+        wt.commit('rev-merged', rev_id='rev-2a',
+                  timestamp=1132586700, timezone=36000,
+                  committer='Joe Foo <joe@foo.com>')
+        wt.set_parent_ids(['rev-1', 'rev-2a'])
+        wt.branch.set_last_revision_info(1, 'rev-1')
+        wt.commit('rev-2', rev_id='rev-2b',
+                  timestamp=1132586800, timezone=36000,
+                  committer='Joe Foo <joe@foo.com>')
+        logfile = self.make_utf8_encoded_stringio()
+        formatter = log.ShortLogFormatter(to_file=logfile)
+        wtb = wt.branch
+        lf = LogCatcher()
+        revspec = revisionspec.RevisionSpec.from_string('1.1.1')
+        rev = revspec.in_history(wtb)
+        self.assertRaises(errors.BzrCommandError, log.show_log, wtb, lf,
+                          start_revision=rev, end_revision=rev)
 
 
 def make_commits_with_trailing_newlines(wt):
-    """Helper method for LogFormatter tests"""    
+    """Helper method for LogFormatter tests"""
     b = wt.branch
     b.nick='test'
     open('a', 'wb').write('hello moto\n')
@@ -290,15 +267,15 @@
     return ''.join(lines)
 
 
-class TestShortLogFormatter(TestCaseWithTransport):
+class TestShortLogFormatter(tests.TestCaseWithTransport):
 
     def test_trailing_newlines(self):
         wt = self.make_branch_and_tree('.')
         b = make_commits_with_trailing_newlines(wt)
         sio = self.make_utf8_encoded_stringio()
-        lf = ShortLogFormatter(to_file=sio)
-        show_log(b, lf)
-        self.assertEqualDiff(sio.getvalue(), """\
+        lf = log.ShortLogFormatter(to_file=sio)
+        log.show_log(b, lf)
+        self.assertEqualDiff("""\
     3 Joe Foo\t2005-11-21
       single line with trailing newline
 
@@ -310,67 +287,66 @@
     1 Joe Foo\t2005-11-21
       simple log message
 
-""")
+""",
+                             sio.getvalue())
 
     def test_short_log_with_merges(self):
         wt = self.make_branch_and_memory_tree('.')
         wt.lock_write()
-        try:
-            wt.add('')
-            wt.commit('rev-1', rev_id='rev-1',
-                      timestamp=1132586655, timezone=36000,
-                      committer='Joe Foo <joe@foo.com>')
-            wt.commit('rev-merged', rev_id='rev-2a',
-                      timestamp=1132586700, timezone=36000,
-                      committer='Joe Foo <joe@foo.com>')
-            wt.set_parent_ids(['rev-1', 'rev-2a'])
-            wt.branch.set_last_revision_info(1, 'rev-1')
-            wt.commit('rev-2', rev_id='rev-2b',
-                      timestamp=1132586800, timezone=36000,
-                      committer='Joe Foo <joe@foo.com>')
-            logfile = self.make_utf8_encoded_stringio()
-            formatter = ShortLogFormatter(to_file=logfile)
-            show_log(wt.branch, formatter)
-            self.assertEqualDiff(logfile.getvalue(), """\
+        self.addCleanup(wt.unlock)
+        wt.add('')
+        wt.commit('rev-1', rev_id='rev-1',
+                  timestamp=1132586655, timezone=36000,
+                  committer='Joe Foo <joe@foo.com>')
+        wt.commit('rev-merged', rev_id='rev-2a',
+                  timestamp=1132586700, timezone=36000,
+                  committer='Joe Foo <joe@foo.com>')
+        wt.set_parent_ids(['rev-1', 'rev-2a'])
+        wt.branch.set_last_revision_info(1, 'rev-1')
+        wt.commit('rev-2', rev_id='rev-2b',
+                  timestamp=1132586800, timezone=36000,
+                  committer='Joe Foo <joe@foo.com>')
+        logfile = self.make_utf8_encoded_stringio()
+        formatter = log.ShortLogFormatter(to_file=logfile)
+        log.show_log(wt.branch, formatter)
+        self.assertEqualDiff("""\
     2 Joe Foo\t2005-11-22 [merge]
       rev-2
 
     1 Joe Foo\t2005-11-22
       rev-1
 
-""")
-        finally:
-            wt.unlock()
+""",
+                             logfile.getvalue())
 
     def test_short_log_single_merge_revision(self):
         wt = self.make_branch_and_memory_tree('.')
         wt.lock_write()
-        try:
-            wt.add('')
-            wt.commit('rev-1', rev_id='rev-1',
-                      timestamp=1132586655, timezone=36000,
-                      committer='Joe Foo <joe@foo.com>')
-            wt.commit('rev-merged', rev_id='rev-2a',
-                      timestamp=1132586700, timezone=36000,
-                      committer='Joe Foo <joe@foo.com>')
-            wt.set_parent_ids(['rev-1', 'rev-2a'])
-            wt.branch.set_last_revision_info(1, 'rev-1')
-            wt.commit('rev-2', rev_id='rev-2b',
-                      timestamp=1132586800, timezone=36000,
-                      committer='Joe Foo <joe@foo.com>')
-            logfile = self.make_utf8_encoded_stringio()
-            formatter = ShortLogFormatter(to_file=logfile)
-            revspec = RevisionSpec.from_string('1.1.1')
-            wtb = wt.branch
-            rev = revspec.in_history(wtb)
-            show_log(wtb, formatter, start_revision=rev, end_revision=rev)
-            self.assertEqualDiff(logfile.getvalue(), """\
+        self.addCleanup(wt.unlock)
+        wt.add('')
+        wt.commit('rev-1', rev_id='rev-1',
+                  timestamp=1132586655, timezone=36000,
+                  committer='Joe Foo <joe@foo.com>')
+        wt.commit('rev-merged', rev_id='rev-2a',
+                  timestamp=1132586700, timezone=36000,
+                  committer='Joe Foo <joe@foo.com>')
+        wt.set_parent_ids(['rev-1', 'rev-2a'])
+        wt.branch.set_last_revision_info(1, 'rev-1')
+        wt.commit('rev-2', rev_id='rev-2b',
+                  timestamp=1132586800, timezone=36000,
+                  committer='Joe Foo <joe@foo.com>')
+        logfile = self.make_utf8_encoded_stringio()
+        formatter = log.ShortLogFormatter(to_file=logfile)
+        revspec = revisionspec.RevisionSpec.from_string('1.1.1')
+        wtb = wt.branch
+        rev = revspec.in_history(wtb)
+        log.show_log(wtb, formatter, start_revision=rev, end_revision=rev)
+        self.assertEqualDiff("""\
 1.1.1 Joe Foo\t2005-11-22
       rev-merged
 
-""")
-        finally:
-            wt.unlock()
+""",
+                             logfile.getvalue())
 
 
 class TestLongLogFormatter(TestCaseWithoutPropsHandler):
@@ -386,17 +362,17 @@
         wt.add('a')
         # XXX: why does a longer nick show up?
         b.nick = 'test_verbose_log'
-        wt.commit(message='add a', 
-                  timestamp=1132711707, 
+        wt.commit(message='add a',
+                  timestamp=1132711707,
                   timezone=36000,
                   committer='Lorem Ipsum <test@example.com>')
         logfile = file('out.tmp', 'w+')
-        formatter = LongLogFormatter(to_file=logfile)
-        show_log(b, formatter, verbose=True)
+        formatter = log.LongLogFormatter(to_file=logfile)
+        log.show_log(b, formatter, verbose=True)
         logfile.flush()
         logfile.seek(0)
         log_contents = logfile.read()
-        self.assertEqualDiff(log_contents, '''\
+        self.assertEqualDiff('''\
 ------------------------------------------------------------
 revno: 1
 committer: Lorem Ipsum <test@example.com>
@@ -406,7 +382,8 @@
   add a
 added:
   a
-''')
+''',
+                             log_contents)
 
     def test_merges_are_indented_by_level(self):
         wt = self.make_branch_and_tree('parent')
@@ -424,10 +401,10 @@
         wt.commit('merge branch 1')
         b = wt.branch
         sio = self.make_utf8_encoded_stringio()
-        lf = LongLogFormatter(to_file=sio)
-        show_log(b, lf, verbose=True)
-        log = normalize_log(sio.getvalue())
-        self.assertEqualDiff(log, """\
+        lf = log.LongLogFormatter(to_file=sio)
+        log.show_log(b, lf, verbose=True)
+        the_log = normalize_log(sio.getvalue())
+        self.assertEqualDiff("""\
 ------------------------------------------------------------
 revno: 2
 committer: Lorem Ipsum <test@example.com>
@@ -463,7 +440,8 @@
 timestamp: Just now
 message:
   first post
-""")
+""",
+                             the_log)
 
     def test_verbose_merge_revisions_contain_deltas(self):
         wt = self.make_branch_and_tree('parent')
@@ -480,10 +458,10 @@
         wt.commit('merge branch 1')
         b = wt.branch
         sio = self.make_utf8_encoded_stringio()
-        lf = LongLogFormatter(to_file=sio)
-        show_log(b, lf, verbose=True)
-        log = normalize_log(sio.getvalue())
-        self.assertEqualDiff(log, """\
+        lf = log.LongLogFormatter(to_file=sio)
+        log.show_log(b, lf, verbose=True)
+        the_log = normalize_log(sio.getvalue())
+        self.assertEqualDiff("""\
 ------------------------------------------------------------
 revno: 2
 committer: Lorem Ipsum <test@example.com>
@@ -516,15 +494,16 @@
 added:
   f1
   f2
-""")
+""",
+                             the_log)
 
     def test_trailing_newlines(self):
         wt = self.make_branch_and_tree('.')
         b = make_commits_with_trailing_newlines(wt)
         sio = self.make_utf8_encoded_stringio()
-        lf = LongLogFormatter(to_file=sio)
-        show_log(b, lf)
-        self.assertEqualDiff(sio.getvalue(), """\
+        lf = log.LongLogFormatter(to_file=sio)
+        log.show_log(b, lf)
+        self.assertEqualDiff("""\
 ------------------------------------------------------------
 revno: 3
 committer: Joe Foo <joe@foo.com>
@@ -549,7 +528,8 @@
 timestamp: Mon 2005-11-21 09:24:15 -0600
 message:
   simple log message
-""")
+""",
+                             sio.getvalue())
 
     def test_author_in_log(self):
         """Log includes the author name if it's set in
@@ -566,9 +546,9 @@
                   committer='Lorem Ipsum <test@example.com>',
                   author='John Doe <jdoe@example.com>')
         sio = StringIO()
-        formatter = LongLogFormatter(to_file=sio)
-        show_log(b, formatter)
-        self.assertEqualDiff(sio.getvalue(), '''\
+        formatter = log.LongLogFormatter(to_file=sio)
+        log.show_log(b, formatter)
+        self.assertEqualDiff('''\
 ------------------------------------------------------------
 revno: 1
 author: John Doe <jdoe@example.com>
@@ -577,7 +557,8 @@
 timestamp: Wed 2005-11-23 12:08:27 +1000
 message:
   add a
-''')
+''',
+                             sio.getvalue())
 
     def test_properties_in_log(self):
         """Log includes the custom properties returned by the registered 
@@ -594,19 +575,19 @@
                   committer='Lorem Ipsum <test@example.com>',
                   author='John Doe <jdoe@example.com>')
         sio = StringIO()
-        formatter = LongLogFormatter(to_file=sio)
+        formatter = log.LongLogFormatter(to_file=sio)
         try:
             def trivial_custom_prop_handler(revision):
                 return {'test_prop':'test_value'}
-            
+
             log.properties_handler_registry.register(
-                'trivial_custom_prop_handler', 
+                'trivial_custom_prop_handler',
                 trivial_custom_prop_handler)
-            show_log(b, formatter)
+            log.show_log(b, formatter)
         finally:
             log.properties_handler_registry.remove(
                 'trivial_custom_prop_handler')
-            self.assertEqualDiff(sio.getvalue(), '''\
+            self.assertEqualDiff('''\
 ------------------------------------------------------------
 revno: 1
 test_prop: test_value
@@ -616,7 +597,8 @@
 timestamp: Wed 2005-11-23 12:08:27 +1000
 message:
   add a
-''')
+''',
+                                 sio.getvalue())
 
     def test_error_in_properties_handler(self):
         """Log includes the custom properties returned by the registered 
@@ -634,19 +616,19 @@
                   author='John Doe <jdoe@example.com>',
                   revprops={'first_prop':'first_value'})
         sio = StringIO()
-        formatter = LongLogFormatter(to_file=sio)
+        formatter = log.LongLogFormatter(to_file=sio)
         try:
             def trivial_custom_prop_handler(revision):
                 raise StandardError("a test error")
-            
+
             log.properties_handler_registry.register(
-                'trivial_custom_prop_handler', 
+                'trivial_custom_prop_handler',
                 trivial_custom_prop_handler)
-            self.assertRaises(StandardError, show_log, b, formatter,)
+            self.assertRaises(StandardError, log.show_log, b, formatter,)
         finally:
             log.properties_handler_registry.remove(
                 'trivial_custom_prop_handler')
-                
+
     def test_properties_handler_bad_argument(self):
         wt = self.make_branch_and_tree('.')
         b = wt.branch
@@ -660,28 +642,28 @@
                   author='John Doe <jdoe@example.com>',
                   revprops={'a_prop':'test_value'})
         sio = StringIO()
-        formatter = LongLogFormatter(to_file=sio)
+        formatter = log.LongLogFormatter(to_file=sio)
         try:
             def bad_argument_prop_handler(revision):
                 return {'custom_prop_name':revision.properties['a_prop']}
-                
+
             log.properties_handler_registry.register(
-                'bad_argument_prop_handler', 
+                'bad_argument_prop_handler',
                 bad_argument_prop_handler)
-            
-            self.assertRaises(AttributeError, formatter.show_properties, 
-                'a revision', '')
-            
+
+            self.assertRaises(AttributeError, formatter.show_properties,
+                              'a revision', '')
+
             revision = b.repository.get_revision(b.last_revision())
             formatter.show_properties(revision, '')
-            self.assertEqualDiff(sio.getvalue(),
-                '''custom_prop_name: test_value\n''')
+            self.assertEqualDiff('''custom_prop_name: test_value\n''',
+                                 sio.getvalue())
         finally:
             log.properties_handler_registry.remove(
                 'bad_argument_prop_handler')
 
 
-class TestLineLogFormatter(TestCaseWithTransport):
+class TestLineLogFormatter(tests.TestCaseWithTransport):
 
     def test_line_log(self):
         """Line log should show revno
@@ -698,57 +680,57 @@
                   timezone=36000,
                   committer='Line-Log-Formatter Tester <test@line.log>')
         logfile = file('out.tmp', 'w+')
-        formatter = LineLogFormatter(to_file=logfile)
-        show_log(b, formatter)
+        formatter = log.LineLogFormatter(to_file=logfile)
+        log.show_log(b, formatter)
         logfile.flush()
         logfile.seek(0)
         log_contents = logfile.read()
-        self.assertEqualDiff(log_contents,
-            '1: Line-Log-Formatte... 2005-11-23 add a\n')
+        self.assertEqualDiff('1: Line-Log-Formatte... 2005-11-23 add a\n',
+                             log_contents)
 
     def test_trailing_newlines(self):
         wt = self.make_branch_and_tree('.')
         b = make_commits_with_trailing_newlines(wt)
         sio = self.make_utf8_encoded_stringio()
-        lf = LineLogFormatter(to_file=sio)
-        show_log(b, lf)
-        self.assertEqualDiff(sio.getvalue(), """\
+        lf = log.LineLogFormatter(to_file=sio)
+        log.show_log(b, lf)
+        self.assertEqualDiff("""\
 3: Joe Foo 2005-11-21 single line with trailing newline
 2: Joe Bar 2005-11-21 multiline
 1: Joe Foo 2005-11-21 simple log message
-""")
+""",
+                             sio.getvalue())
 
     def test_line_log_single_merge_revision(self):
         wt = self.make_branch_and_memory_tree('.')
         wt.lock_write()
-        try:
-            wt.add('')
-            wt.commit('rev-1', rev_id='rev-1',
-                      timestamp=1132586655, timezone=36000,
-                      committer='Joe Foo <joe@foo.com>')
-            wt.commit('rev-merged', rev_id='rev-2a',
-                      timestamp=1132586700, timezone=36000,
-                      committer='Joe Foo <joe@foo.com>')
-            wt.set_parent_ids(['rev-1', 'rev-2a'])
-            wt.branch.set_last_revision_info(1, 'rev-1')
-            wt.commit('rev-2', rev_id='rev-2b',
-                      timestamp=1132586800, timezone=36000,
-                      committer='Joe Foo <joe@foo.com>')
-            logfile = self.make_utf8_encoded_stringio()
-            formatter = LineLogFormatter(to_file=logfile)
-            revspec = RevisionSpec.from_string('1.1.1')
-            wtb = wt.branch
-            rev = revspec.in_history(wtb)
-            show_log(wtb, formatter, start_revision=rev, end_revision=rev)
-            self.assertEqualDiff(logfile.getvalue(), """\
+        self.addCleanup(wt.unlock)
+        wt.add('')
+        wt.commit('rev-1', rev_id='rev-1',
+                  timestamp=1132586655, timezone=36000,
+                  committer='Joe Foo <joe@foo.com>')
+        wt.commit('rev-merged', rev_id='rev-2a',
+                  timestamp=1132586700, timezone=36000,
+                  committer='Joe Foo <joe@foo.com>')
+        wt.set_parent_ids(['rev-1', 'rev-2a'])
+        wt.branch.set_last_revision_info(1, 'rev-1')
+        wt.commit('rev-2', rev_id='rev-2b',
+                  timestamp=1132586800, timezone=36000,
+                  committer='Joe Foo <joe@foo.com>')
+        logfile = self.make_utf8_encoded_stringio()
+        formatter = log.LineLogFormatter(to_file=logfile)
+        revspec = revisionspec.RevisionSpec.from_string('1.1.1')
+        wtb = wt.branch
+        rev = revspec.in_history(wtb)
+        log.show_log(wtb, formatter, start_revision=rev, end_revision=rev)
+        self.assertEqualDiff("""\
 1.1.1: Joe Foo 2005-11-22 rev-merged
-""")
-        finally:
-            wt.unlock()
-
-
-
-class TestGetViewRevisions(TestCaseWithTransport):
+""",
+                             logfile.getvalue())
+
+
+
+class TestGetViewRevisions(tests.TestCaseWithTransport):
 
     def make_tree_with_commits(self):
         """Create a tree with well-known revision ids"""
@@ -775,18 +757,26 @@
     def make_tree_with_many_merges(self):
         """Create a tree with well-known revision ids"""
         wt = self.make_branch_and_tree('tree1')
+        self.build_tree_contents([('tree1/f', '1\n')])
+        wt.add(['f'], ['f-id'])
         wt.commit('commit one', rev_id='1')
         wt.commit('commit two', rev_id='2')
+
         tree3 = wt.bzrdir.sprout('tree3').open_workingtree()
+        self.build_tree_contents([('tree3/f', '1\n2\n3a\n')])
         tree3.commit('commit three a', rev_id='3a')
+
         tree2 = wt.bzrdir.sprout('tree2').open_workingtree()
         tree2.merge_from_branch(tree3.branch)
         tree2.commit('commit three b', rev_id='3b')
+
         wt.merge_from_branch(tree2.branch)
         wt.commit('commit three c', rev_id='3c')
         tree2.commit('four-a', rev_id='4a')
+
         wt.merge_from_branch(tree2.branch)
         wt.commit('four-b', rev_id='4b')
+
         mainline_revs = [None, '1', '2', '3c', '4b']
         rev_nos = {'1':1, '2':2, '3c': 3, '4b':4}
         full_rev_nos_for_reference = {
@@ -805,12 +795,13 @@
         mainline_revs, rev_nos, wt = self.make_tree_with_commits()
         wt.lock_read()
         self.addCleanup(wt.unlock)
-        revisions = list(get_view_revisions(mainline_revs, rev_nos, wt.branch,
-                                            'forward'))
+        revisions = list(log.get_view_revisions(
+                mainline_revs, rev_nos, wt.branch, 'forward'))
         self.assertEqual([('1', '1', 0), ('2', '2', 0), ('3', '3', 0)],
-            revisions)
-        revisions2 = list(get_view_revisions(mainline_revs, rev_nos, wt.branch,
-                                             'forward', include_merges=False))
+                         revisions)
+        revisions2 = list(log.get_view_revisions(
+                mainline_revs, rev_nos, wt.branch, 'forward',
+                include_merges=False))
         self.assertEqual(revisions, revisions2)
 
     def test_get_view_revisions_reverse(self):
@@ -818,12 +809,13 @@
         mainline_revs, rev_nos, wt = self.make_tree_with_commits()
         wt.lock_read()
         self.addCleanup(wt.unlock)
-        revisions = list(get_view_revisions(mainline_revs, rev_nos, wt.branch,
-                                            'reverse'))
+        revisions = list(log.get_view_revisions(
+                mainline_revs, rev_nos, wt.branch, 'reverse'))
         self.assertEqual([('3', '3', 0), ('2', '2', 0), ('1', '1', 0), ],
-            revisions)
-        revisions2 = list(get_view_revisions(mainline_revs, rev_nos, wt.branch,
-                                             'reverse', include_merges=False))
+                         revisions)
+        revisions2 = list(log.get_view_revisions(
+                mainline_revs, rev_nos, wt.branch, 'reverse',
+                include_merges=False))
         self.assertEqual(revisions, revisions2)
 
     def test_get_view_revisions_merge(self):
@@ -831,52 +823,88 @@
         mainline_revs, rev_nos, wt = self.make_tree_with_merges()
         wt.lock_read()
         self.addCleanup(wt.unlock)
-        revisions = list(get_view_revisions(mainline_revs, rev_nos, wt.branch,
-                                            'forward'))
-        self.assertEqual([('1', '1', 0), ('2', '2', 0), ('3', '3', 0),
-            ('4b', '4', 0), ('4a', '3.1.1', 1)],
-            revisions)
-        revisions = list(get_view_revisions(mainline_revs, rev_nos, wt.branch,
-                                             'forward', include_merges=False))
-        self.assertEqual([('1', '1', 0), ('2', '2', 0), ('3', '3', 0),
-            ('4b', '4', 0)],
-            revisions)
+        revisions = list(log.get_view_revisions(
+                mainline_revs, rev_nos, wt.branch, 'forward'))
+        self.assertEqual([('1', '1', 0), ('2', '2', 0), ('3', '3', 0),
+                          ('4b', '4', 0), ('4a', '3.1.1', 1)],
+                         revisions)
+        revisions = list(log.get_view_revisions(
+                mainline_revs, rev_nos, wt.branch, 'forward',
+                include_merges=False))
+        self.assertEqual([('1', '1', 0), ('2', '2', 0), ('3', '3', 0),
+                          ('4b', '4', 0)],
+                         revisions)
 
     def test_get_view_revisions_merge_reverse(self):
         """Test get_view_revisions in reverse when there are merges"""
         mainline_revs, rev_nos, wt = self.make_tree_with_merges()
         wt.lock_read()
         self.addCleanup(wt.unlock)
-        revisions = list(get_view_revisions(mainline_revs, rev_nos, wt.branch,
-                                            'reverse'))
+        revisions = list(log.get_view_revisions(
+                mainline_revs, rev_nos, wt.branch, 'reverse'))
         self.assertEqual([('4b', '4', 0), ('4a', '3.1.1', 1),
-            ('3', '3', 0), ('2', '2', 0), ('1', '1', 0)],
-            revisions)
-        revisions = list(get_view_revisions(mainline_revs, rev_nos, wt.branch,
-                                             'reverse', include_merges=False))
+                          ('3', '3', 0), ('2', '2', 0), ('1', '1', 0)],
+                         revisions)
+        revisions = list(log.get_view_revisions(
+                mainline_revs, rev_nos, wt.branch, 'reverse',
+                include_merges=False))
         self.assertEqual([('4b', '4', 0), ('3', '3', 0), ('2', '2', 0),
-            ('1', '1', 0)],
-            revisions)
+                          ('1', '1', 0)],
+                         revisions)
 
     def test_get_view_revisions_merge2(self):
         """Test get_view_revisions when there are merges"""
         mainline_revs, rev_nos, wt = self.make_tree_with_many_merges()
         wt.lock_read()
         self.addCleanup(wt.unlock)
-        revisions = list(get_view_revisions(mainline_revs, rev_nos, wt.branch,
-                                            'forward'))
+        revisions = list(log.get_view_revisions(
+                mainline_revs, rev_nos, wt.branch, 'forward'))
         expected = [('1', '1', 0), ('2', '2', 0), ('3c', '3', 0),
-            ('3a', '2.1.1', 1), ('3b', '2.2.1', 1), ('4b', '4', 0),
-            ('4a', '2.2.2', 1)]
+                    ('3a', '2.1.1', 1), ('3b', '2.2.1', 1), ('4b', '4', 0),
+                    ('4a', '2.2.2', 1)]
         self.assertEqual(expected, revisions)
-        revisions = list(get_view_revisions(mainline_revs, rev_nos, wt.branch,
-                                             'forward', include_merges=False))
+        revisions = list(log.get_view_revisions(
+                mainline_revs, rev_nos, wt.branch, 'forward',
+                include_merges=False))
         self.assertEqual([('1', '1', 0), ('2', '2', 0), ('3c', '3', 0),
-            ('4b', '4', 0)],
-            revisions)
-
-
-class TestGetRevisionsTouchingFileID(TestCaseWithTransport):
+                          ('4b', '4', 0)],
+                         revisions)
+
+
+    def test_file_id_for_range(self):
+        mainline_revs, rev_nos, wt = self.make_tree_with_many_merges()
+        wt.lock_read()
+        self.addCleanup(wt.unlock)
+
+        def rev_from_rev_id(revid, branch):
+            revspec = revisionspec.RevisionSpec.from_string('revid:%s' % revid)
+            return revspec.in_history(branch)
+
+        def view_revs(start_rev, end_rev, file_id, direction):
+            revs = log.calculate_view_revisions(
+                wt.branch,
+                start_rev, # start_revision
+                end_rev, # end_revision
+                direction, # direction
+                file_id, # specific_fileid
+                True, # generate_merge_revisions
+                True, # allow_single_merge_revision
+                )
+            return revs
+
+        rev_3a = rev_from_rev_id('3a', wt.branch)
+        rev_4b = rev_from_rev_id('4b', wt.branch)
+        self.assertEqual([('3c', '3', 0), ('3a', '2.1.1', 1)],
+                          view_revs(rev_3a, rev_4b, 'f-id', 'reverse'))
+        # Note that the depth is 0 for 3a because depths are normalized, but
+        # there is still a bug somewhere... most probably in
+        # _filter_revision_range and/or get_view_revisions still around a bad
+        # use of reverse_by_depth
+        self.assertEqual([('3a', '2.1.1', 0)],
+                          view_revs(rev_3a, rev_4b, 'f-id', 'forward'))
+
+
+class TestGetRevisionsTouchingFileID(tests.TestCaseWithTransport):
 
     def create_tree_with_single_merge(self):
         """Create a branch with a moderate layout.
@@ -932,54 +960,51 @@
         tree.commit('D', rev_id='D')
 
         # Switch to a read lock for this tree.
-        # We still have addCleanup(unlock)
+        # We still have an addCleanup(tree.unlock) pending
         tree.unlock()
         tree.lock_read()
         return tree
 
+    def check_delta(self, delta, **kw):
+        """Check the filenames touched by a delta are as expected.
+
+        Caller only have to pass in the list of files for each part, all
+        unspecified parts are considered empty (and checked as such).
+        """
+        for n in 'added', 'removed', 'renamed', 'modified', 'unchanged':
+            # By default we expect an empty list
+            expected = kw.get(n, [])
+            # strip out only the path components
+            got = [x[0] for x in getattr(delta, n)]
+            self.assertEqual(expected, got)
+
     def test_tree_with_single_merge(self):
         """Make sure the tree layout is correct."""
         tree = self.create_tree_with_single_merge()
         rev_A_tree = tree.branch.repository.revision_tree('A')
         rev_B_tree = tree.branch.repository.revision_tree('B')
-
-        f1_changed = (u'f1', 'f1-id', 'file', True, False)
-        f2_changed = (u'f2', 'f2-id', 'file', True, False)
-        f3_changed = (u'f3', 'f3-id', 'file', True, False)
-
-        delta = rev_B_tree.changes_from(rev_A_tree)
-        self.assertEqual([f1_changed, f3_changed], delta.modified)
-        self.assertEqual([], delta.renamed)
-        self.assertEqual([], delta.added)
-        self.assertEqual([], delta.removed)
-
         rev_C_tree = tree.branch.repository.revision_tree('C')
-        delta = rev_C_tree.changes_from(rev_A_tree)
-        self.assertEqual([f2_changed, f3_changed], delta.modified)
-        self.assertEqual([], delta.renamed)
-        self.assertEqual([], delta.added)
-        self.assertEqual([], delta.removed)
-
         rev_D_tree = tree.branch.repository.revision_tree('D')
-        delta = rev_D_tree.changes_from(rev_B_tree)
-        self.assertEqual([f2_changed, f3_changed], delta.modified)
-        self.assertEqual([], delta.renamed)
-        self.assertEqual([], delta.added)
-        self.assertEqual([], delta.removed)
-
-        delta = rev_D_tree.changes_from(rev_C_tree)
-        self.assertEqual([f1_changed, f3_changed], delta.modified)
-        self.assertEqual([], delta.renamed)
-        self.assertEqual([], delta.added)
-        self.assertEqual([], delta.removed)
+
+        self.check_delta(rev_B_tree.changes_from(rev_A_tree),
+                         modified=['f1', 'f3'])
+
+        self.check_delta(rev_C_tree.changes_from(rev_A_tree),
+                         modified=['f2', 'f3'])
+
+        self.check_delta(rev_D_tree.changes_from(rev_B_tree),
+                         modified=['f2', 'f3'])
+
+        self.check_delta(rev_D_tree.changes_from(rev_C_tree),
+                         modified=['f1', 'f3'])
 
     def assertAllRevisionsForFileID(self, tree, file_id, revisions):
-        """Make sure _filter_revisions_touching_file_id returns the right values.
+        """Ensure _filter_revisions_touching_file_id returns the right values.
 
         Get the return value from _filter_revisions_touching_file_id and make
         sure they are correct.
         """
-        # The api for _get_revisions_touching_file_id is a little crazy,
+        # The api for _filter_revisions_touching_file_id is a little crazy.
         # So we do the setup here.
         mainline = tree.branch.revision_history()
         mainline.insert(0, None)
@@ -989,8 +1014,7 @@
         actual_revs = log._filter_revisions_touching_file_id(
                             tree.branch,
                             file_id,
-                            list(view_revs_iter),
-                            'reverse')
+                            list(view_revs_iter))
         self.assertEqual(revisions, [r for r, revno, depth in actual_revs])
 
     def test_file_id_f1(self):
@@ -1007,13 +1031,14 @@
     def test_file_id_f3(self):
         tree = self.create_tree_with_single_merge()
         # f3 should be marked as modified by revisions A, B, C, and D
-        self.assertAllRevisionsForFileID(tree, 'f2-id', ['D', 'C', 'A'])
+        self.assertAllRevisionsForFileID(tree, 'f3-id', ['D', 'C', 'B', 'A'])
 
     def test_file_id_with_ghosts(self):
         # This is testing bug #209948, where having a ghost would cause
         # _filter_revisions_touching_file_id() to fail.
         tree = self.create_tree_with_single_merge()
         # We need to add a revision, so switch back to a write-locked tree
+        # (still a single addCleanup(tree.unlock) pending).
         tree.unlock()
         tree.lock_write()
         first_parent = tree.last_revision()
@@ -1024,7 +1049,7 @@
         self.assertAllRevisionsForFileID(tree, 'f2-id', ['D', 'C', 'A'])
 
 
-class TestShowChangedRevisions(TestCaseWithTransport):
+class TestShowChangedRevisions(tests.TestCaseWithTransport):
 
     def test_show_changed_revisions_verbose(self):
         tree = self.make_branch_and_tree('tree_a')
@@ -1037,12 +1062,12 @@
         self.assertNotContainsRe(s.getvalue(), 'foo')
 
 
-class TestLogFormatter(TestCase):
+class TestLogFormatter(tests.TestCase):
 
     def test_short_committer(self):
-        rev = Revision('a-id')
+        rev = revision.Revision('a-id')
         rev.committer = 'John Doe <jdoe@example.com>'
-        lf = LogFormatter(None)
+        lf = log.LogFormatter(None)
         self.assertEqual('John Doe', lf.short_committer(rev))
         rev.committer = 'John Smith <jsmith@example.com>'
         self.assertEqual('John Smith', lf.short_committer(rev))
@@ -1056,9 +1081,9 @@
         self.assertEqual('John Smith', lf.short_committer(rev))
 
     def test_short_author(self):
-        rev = Revision('a-id')
+        rev = revision.Revision('a-id')
         rev.committer = 'John Doe <jdoe@example.com>'
-        lf = LogFormatter(None)
+        lf = log.LogFormatter(None)
         self.assertEqual('John Doe', lf.short_author(rev))
         rev.properties['author'] = 'John Smith <jsmith@example.com>'
         self.assertEqual('John Smith', lf.short_author(rev))
@@ -1070,3 +1095,57 @@
         self.assertEqual('jsmith@example.com', lf.short_author(rev))
         rev.properties['author'] = 'John Smith jsmith@example.com'
         self.assertEqual('John Smith', lf.short_author(rev))
+
+
+class TestReverseByDepth(tests.TestCase):
+    """Test reverse_by_depth behavior.
+
+    This is used to present revisions in forward (oldest first) order in a nice
+    layout.
+
+    The tests use lighter revision description to ease reading.
+    """
+
+    def assertReversed(self, forward, backward):
+        # Transform the descriptions to suit the API: tests use (revno, depth),
+        # while the API expects (revid, revno, depth)
+        def complete_revisions(l):
+            """Transform the description to suit the API.
+
+            Tests use (revno, depth) whil the API expects (revid, revno, depth).
+            Since the revid is arbitrary, we just duplicate revno
+            """
+            return [ (r, r, d) for r, d in l]
+        forward = complete_revisions(forward)
+        backward= complete_revisions(backward)
+        self.assertEqual(forward, log.reverse_by_depth(backward))
+
+
+    def test_mainline_revisions(self):
+        self.assertReversed([( '1', 0), ('2', 0)],
+                            [('2', 0), ('1', 0)])
+
+    def test_merged_revisions(self):
+        self.assertReversed([('1', 0), ('2', 0), ('2.2', 1), ('2.1', 1),],
+                            [('2', 0), ('2.1', 1), ('2.2', 1), ('1', 0),])
+    def test_shifted_merged_revisions(self):
+        """Test irregular layout.
+
+        Requesting revisions touching a file can produce "holes" in the depths.
+        """
+        self.assertReversed([('1', 0), ('2', 0), ('1.1', 2), ('1.2', 2),],
+                            [('2', 0), ('1.2', 2), ('1.1', 2), ('1', 0),])
+
+    def test_merged_without_child_revisions(self):
+        """Test irregular layout.
+
+        Revision ranges can produce "holes" in the depths.
+        """
+        # When a revision of higher depth doesn't follow one of lower depth, we
+        # assume a lower depth one is virtually there
+        self.assertReversed([('1', 2), ('2', 2), ('3', 3), ('4', 4)],
+                            [('4', 4), ('3', 3), ('2', 2), ('1', 2),])
+        # So we get the same order after reversing below even if the original
+        # revisions are not in the same order.
+        self.assertReversed([('1', 2), ('2', 2), ('3', 3), ('4', 4)],
+                            [('3', 3), ('4', 4), ('2', 2), ('1', 2),])

